@@ -4,7 +4,8 @@ import datetime
 import decimal
 import uuid
 from collections import OrderedDict
-from typing import Any, Dict, Generic, Mapping, List, Tuple, Type, TypeVar
+from enum import Enum
+from typing import Any, Dict, Generic, Iterable, Mapping, Tuple, Type, TypeVar
 
 import rest_framework.fields
 import rest_framework.serializers
@@ -21,11 +22,31 @@ from rest_framework_dataclasses.types import Dataclass
 
 
 # Define some types to make type hinting more readable
-AnyT = TypeVar('AnyT')
 KWArgs = Dict[str, Any]
 SerializerField = rest_framework.fields.Field
 SerializerFieldDefinition = Tuple[Type[SerializerField], KWArgs]
 T = TypeVar('T', bound=Dataclass)
+AnyT = TypeVar('AnyT')
+
+
+# Helper function to strip the empty sentinel value and replace it with the default value from a dataclass
+def _strip_empty_sentinels(data: AnyT, instance: AnyT = None) -> AnyT:
+    if dataclasses.is_dataclass(data) and not isinstance(data, type):
+        values = {field.name: _strip_empty_sentinels(getattr(data, field.name),
+                                                     getattr(instance, field.name, None))
+                  for field in dataclasses.fields(data)
+                  if getattr(data, field.name) is not empty}
+        if instance:
+            for field, value in values.items():
+                setattr(instance, field, value)
+            return instance
+        else:
+            return type(data)(**values)
+    elif isinstance(data, list):
+        return [_strip_empty_sentinels(item) for item in data]
+    elif isinstance(data, dict):
+        return {key: _strip_empty_sentinels(value) for key, value in data.items()}
+    return data
 
 
 # noinspection PyMethodMayBeStatic
@@ -42,8 +63,6 @@ class DataclassSerializer(rest_framework.serializers.Serializer, Generic[T]):
 
     If the `DataclassSerializer` class *doesn't* generate the set of fields that you need you should either declare the
     extra/differing fields explicitly on the serializer class, or simply use a `Serializer` class.
-
-    Note that this implementation is heavily based on that of ModelSerializer in the rest_framework source code.
     """
 
     # The mapping of field types to serializer fields
@@ -58,13 +77,20 @@ class DataclassSerializer(rest_framework.serializers.Serializer, Generic[T]):
         datetime.time:      rest_framework.fields.TimeField,
         datetime.timedelta: rest_framework.fields.DurationField,
         uuid.UUID:          rest_framework.fields.UUIDField,
-        # Note that these are only used for class typehints, typing.Dict and typing.List are handled by
-        # build_composite_field().
         dict:               rest_framework.fields.DictField,
         list:               rest_framework.fields.ListField
     }
     serializer_related_field = PrimaryKeyRelatedField
 
+    # Unfortunately this cannot be an actual field as Python processes the class before it defines the class, but this
+    # comes close enough.
+    @property
+    def serializer_dataclass_field(self):
+        return DataclassSerializer
+
+    # Type hints
+    _declared_fields: Mapping[str, SerializerField]
+
     # Override constructor to allow "anonymous" usage by passing the dataclass type and extra kwargs as a constructor
     # parameter instead of via a Meta class.
     def __init__(self, *args, **kwargs):
@@ -72,6 +98,23 @@ class DataclassSerializer(rest_framework.serializers.Serializer, Generic[T]):
         self.extra_kwargs = kwargs.pop('extra_kwargs', None)
         super(DataclassSerializer, self).__init__(*args, **kwargs)
 
+    @classmethod
+    def many_init(cls, *args, **kwargs):
+        """
+        Implements the creation of a `DataclassListSerializer` parent class when `many=True` is used.
+        """
+        list_kwargs = {key: value for key, value in kwargs.items()
+                       if key in rest_framework.serializers.LIST_SERIALIZER_KWARGS}
+        list_kwargs['child'] = cls(*args, **kwargs)
+
+        allow_empty = kwargs.pop('allow_empty', None)
+        if allow_empty is not None:
+            list_kwargs['allow_empty'] = allow_empty
+
+        meta = getattr(cls, 'Meta', None)
+        list_serializer_class = getattr(meta, 'list_serializer_class', DataclassListSerializer)
+        return list_serializer_class(*args, **list_kwargs)
+
     # Parse and validate the configuration to a more usable format
 
     @cached_property
@@ -89,12 +132,14 @@ class DataclassSerializer(rest_framework.serializers.Serializer, Generic[T]):
                 "Class '{serializer_class}' missing `Meta` attribute."
                 .format(serializer_class=self.__class__.__name__)
             )
-            assert hasattr(self.Meta, 'dataclass'), (
+
+            meta = getattr(self, 'Meta')
+            assert hasattr(meta, 'dataclass'), (
                 "Class '{serializer_class}' missing `Meta.dataclass` attribute."
                 .format(serializer_class=self.__class__.__name__)
             )
 
-            dataclass_type = self.Meta.dataclass
+            dataclass_type = meta.dataclass
 
         # Make sure we're dealing with an actual dataclass.
         if not dataclasses.is_dataclass(dataclass_type):
@@ -107,32 +152,11 @@ class DataclassSerializer(rest_framework.serializers.Serializer, Generic[T]):
 
     # Default `create` and `update` behavior...
 
-    def _strip_empty_sentinels(self, x: AnyT) -> AnyT:
-        """
-        Recursively strip the `empty` sentinel values from dataclasses, replacing them with their fields default value.
-        """
-        if dataclasses.is_dataclass(x) and not isinstance(x, type):
-            values = {field.name: self._strip_empty_sentinels(getattr(x, field.name)) for field in dataclasses.fields(x)
-                      if getattr(x, field.name) is not empty}
-            return type(x)(**values)
-        if isinstance(x, list):
-            return [self._strip_empty_sentinels(item) for item in x]
-        if isinstance(x, dict):
-            return {key: self._strip_empty_sentinels(value) for key, value in x.items()}
-        return x
-
     def create(self, validated_data: T) -> T:
-        # Create a new instance with all the `empty` sentinel values replaced by their defaults.
-        return self._strip_empty_sentinels(validated_data)
+        return _strip_empty_sentinels(validated_data)
 
     def update(self, instance: T, validated_data: T) -> T:
-        for name in self.dataclass_definition.fields.keys():
-            # Only update fields that have been set in the serialized representation.
-            value = getattr(validated_data, name)
-            if value is not empty:
-                setattr(instance, name, self._strip_empty_sentinels(value))
-
-        return instance
+        return _strip_empty_sentinels(validated_data, instance)
 
     def save(self, **kwargs) -> T:
         assert hasattr(self, '_errors'), (
@@ -171,8 +195,15 @@ class DataclassSerializer(rest_framework.serializers.Serializer, Generic[T]):
         Return the dict of field names -> field instances that should be used for `self.fields` when instantiating the
         serializer.
         """
+        # Copy declared fields, so that the field on the serializer class remains unchanged when the fields are bound to
+        # the serializer instance.
         declared_fields = copy.deepcopy(self._declared_fields)
 
+        # Copy fields declared in metadata, for the same reason.
+        metadata_fields = {field.name: copy.deepcopy(field.metadata['serializer_field'])
+                           for field in self.dataclass_definition.fields.values()
+                           if 'serializer_field' in field.metadata}
+
         # Determine all fields that should be included on the serializer.
         field_names = self.get_field_names()
 
@@ -183,15 +214,16 @@ class DataclassSerializer(rest_framework.serializers.Serializer, Generic[T]):
         fields = OrderedDict()
 
         for field_name in field_names:
-            # If the field is explicitly declared on the class then use that.
-            if field_name in declared_fields:
+            # If the field is explicitly declared, either on the serializer or in the field metadata, then use that.
+            if field_name in declared_fields or field_name in metadata_fields:
                 assert field_name not in extra_kwargs, (
                     "Cannot both declare the field '{field_name}' and include it in '{serializer_class}' "
                     "`extra_kwargs` or `read_only_fields` option. Move all options to the field declaration."
                     .format(field_name=field_name, serializer_class=self.__class__.__name__)
                 )
 
-                fields[field_name] = declared_fields[field_name]
+                # Prefer the field declared on the serializer above the one declared in field metadata.
+                fields[field_name] = declared_fields.get(field_name, metadata_fields.get(field_name, None))
                 continue
 
             extra_field_kwargs = extra_kwargs.get(field_name, {})
@@ -201,14 +233,14 @@ class DataclassSerializer(rest_framework.serializers.Serializer, Generic[T]):
 
     # Methods for determining the set of field names to include...
 
-    def get_field_names(self) -> List[str]:
+    def get_field_names(self) -> Iterable[str]:
         """
         Returns the list of all field names that should be created when instantiating this serializer class. This is
         based on the default set of fields, but also takes into account the `Meta.fields` or `Meta.exclude` options
         if they have been specified.
         """
-        # Retrieve metadata about the declared fields.
-        declared_fields = self._declared_fields
+        # Retrieve names of the declared fields.
+        declared_field_names = self._declared_fields.keys()
 
         # Read configuration from Meta class.
         meta = getattr(self, 'Meta', None)
@@ -242,7 +274,7 @@ class DataclassSerializer(rest_framework.serializers.Serializer, Generic[T]):
             assert (
                 field_name == rest_framework.serializers.ALL_FIELDS or                         # all fields magic option
                 field_name in self.dataclass_definition.fields or                              # dataclass fields
-                field_name in declared_fields or                                               # declared fields
+                field_name in declared_field_names or                                          # declared fields
                 callable(getattr(self.dataclass_definition.dataclass_type, field_name, None))  # methods
             ), (
                 "The field '{field_name}' was included on serializer {serializer_class} in the `fields` option, "
@@ -254,7 +286,7 @@ class DataclassSerializer(rest_framework.serializers.Serializer, Generic[T]):
             # If there are only explicitly specified fields, ensure that all declared fields are included. Do not
             # require any fields that are declared in a parent class, in order to allow serializer subclasses to only
             # include a subset of fields.
-            required_field_names = set(declared_fields)
+            required_field_names = set(declared_field_names)
             for cls in self.__class__.__bases__:
                 required_field_names -= set(getattr(cls, '_declared_fields', []))
 
@@ -269,12 +301,12 @@ class DataclassSerializer(rest_framework.serializers.Serializer, Generic[T]):
         # The field list now includes the magic all fields option, so replace it with the default field names.
         fields = list(fields)
         fields.remove(rest_framework.serializers.ALL_FIELDS)
-        fields.extend(self.get_default_field_names(declared_fields))
+        fields.extend(self.get_default_field_names(declared_field_names))
 
         if exclude is not None:
             # If `Meta.exclude` is included, then remove those fields.
             for field_name in exclude:
-                assert field_name not in declared_fields, (
+                assert field_name not in declared_field_names, (
                     "Cannot both declare the field '{field_name}' and include it in '{serializer_class}' `exclude` "
                     "option. Remove the field or, if inherited from a parent serializer, disable with "
                     "`{field_name} = None`."
@@ -291,7 +323,7 @@ class DataclassSerializer(rest_framework.serializers.Serializer, Generic[T]):
 
         return fields
 
-    def get_default_field_names(self, declared_fields: Mapping[str, SerializerField]) -> List[str]:
+    def get_default_field_names(self, declared_fields: Iterable[str]) -> Iterable[str]:
         """
         Return the default list of field names that will be used if the `Meta.fields` option is not specified.
         """
@@ -311,10 +343,13 @@ class DataclassSerializer(rest_framework.serializers.Serializer, Generic[T]):
 
         # Determine the serializer field class and keyword arguments.
         if source in self.dataclass_definition.fields:
-            default_value = self.dataclass_definition.fields[source].default
-            default_type = type(default_value) if default_value is not None else None
-            type_info = field_utils.get_type_info(self.dataclass_definition.field_types[source], default_type)
+            field = self.dataclass_definition.fields[source]
+            type_info = field_utils.get_type_info(self.dataclass_definition.field_types[source])
             field_class, field_kwargs = self.build_typed_field(source, type_info, extra_kwargs)
+
+            # Include extra kwargs defined in the field metadata
+            if 'serializer_kwargs' in field.metadata:
+                field_kwargs = self.include_extra_kwargs(field_kwargs, field.metadata['serializer_kwargs'])
         elif hasattr(self.dataclass_definition.dataclass_type, source):
             field_class, field_kwargs = self.build_property_field(source)
         else:
@@ -334,18 +369,36 @@ class DataclassSerializer(rest_framework.serializers.Serializer, Generic[T]):
         if type_info.is_mapping or type_info.is_many:
             field_class, field_kwargs = self.build_composite_field(field_name, type_info, extra_kwargs)
         elif dataclasses.is_dataclass(type_info.base_type):
-            field_class, field_kwargs = self.build_nested_field(field_name, type_info)
+            field_class, field_kwargs = self.build_dataclass_field(field_name, type_info)
         elif isinstance(type_info.base_type, type) and issubclass(type_info.base_type, Model):
             field_class, field_kwargs = self.build_relational_field(field_name, type_info)
+        elif isinstance(type_info.base_type, type) and issubclass(type_info.base_type, Enum):
+            field_class, field_kwargs = self.build_enum_field(field_name, type_info)
         elif typing_utils.is_literal_type(type_info.base_type):
             field_class, field_kwargs = self.build_literal_field(field_name, type_info)
         else:
             field_class, field_kwargs = self.build_standard_field(field_name, type_info)
 
-        if type_info.is_optional:
+        # Mark a field as not-required if it has a default value (factory) on the dataclass. This is consistent with the
+        # constructor of dataclasses, where these fields are also made optional. Note that this is different from the
+        # `typing.Optional[]` qualifier, which merely makes the field nullable, but still requires it to be passed. Of
+        # course it makes sense for `Optional` fields to have `None` as a default value, but that's up to the user.
+        field = self.dataclass_definition.fields[field_name]
+        if field.default is not dataclasses.MISSING or field.default_factory is not dataclasses.MISSING:
             field_kwargs['required'] = False
+
+            # Explicitly don't set the default argument here. Setting it would cause the default value to be inserted in
+            # the native representation (`to_internal_value()` argument) if the field wasn't supplied by the user (for
+            # non-partial updates). This in turn would cause `update()` to overwrite non-supplied fields with the
+            # defaults, which is undesirable. Instead, let the dataclass constructor apply the default values when the
+            # dataclass is instantiated.
+
+        # Mark a field as nullable if it is declared as Optional[] (which has a confusing name).
+        if type_info.is_nullable:
             field_kwargs['allow_null'] = True
 
+        # The final qualifier declares that a variable or attribute should not be reassigned (PEP 591). Mark the field
+        # as read only.
         if type_info.is_final:
             field_kwargs['read_only'] = True
 
@@ -356,10 +409,11 @@ class DataclassSerializer(rest_framework.serializers.Serializer, Generic[T]):
         """
         Create a composite (mapping or list) field.
         """
+        # Lookup the types from the field mapping, so that it can easily be changed without overriding the method.
         if type_info.is_mapping:
-            field_class = rest_framework.fields.DictField
+            field_class = self.serializer_field_mapping[dict]
         else:
-            field_class = rest_framework.fields.ListField
+            field_class = self.serializer_field_mapping[list]
 
         # If the base type is not specified or is any type, we don't have to bother creating the child field.
         if type_info.base_type is None:
@@ -414,6 +468,17 @@ class DataclassSerializer(rest_framework.serializers.Serializer, Generic[T]):
 
         return field_class, field_kwargs
 
+    def build_enum_field(self, field_name: str, type_info: TypeInfo) -> SerializerFieldDefinition:
+        """
+        Create EnumField from a Enum type.
+        """
+        field_class = fields.EnumField
+        field_kwargs = {
+            'enum_class': type_info.base_type
+        }
+        return field_class, field_kwargs
+
+    # noinspection PyUnusedLocal
     def build_literal_field(self, field_name: str, type_info: TypeInfo) -> SerializerFieldDefinition:
         """
         Create ChoiceField from a Literal[...] type.
@@ -427,16 +492,21 @@ class DataclassSerializer(rest_framework.serializers.Serializer, Generic[T]):
 
         return field_class, field_kwargs
 
-    def build_nested_field(self, field_name: str, type_info: TypeInfo) -> SerializerFieldDefinition:
+    def build_dataclass_field(self, field_name: str, type_info: TypeInfo) -> SerializerFieldDefinition:
         """
         Create fields for dataclasses.
         """
-        field_class = DataclassSerializer
+        try:
+            field_class = field_utils.lookup_type_in_mapping(self.serializer_field_mapping, type_info.base_type)
+        except KeyError:
+            field_class = self.serializer_dataclass_field
+
         field_kwargs = {'dataclass': type_info.base_type,
                         'many': type_info.is_many}
 
         return field_class, field_kwargs
 
+    # noinspection PyUnusedLocal
     def build_property_field(self, field_name: str) -> SerializerFieldDefinition:
         """
         Create a read only field for dataclass methods and properties.
@@ -502,14 +572,13 @@ class DataclassSerializer(rest_framework.serializers.Serializer, Generic[T]):
                     "The `read_only_fields` option must be a list or tuple. Got '{type}'."
                     .format(type=type(read_only_fields).__name__)
                 )
+
             for field_name in read_only_fields:
                 kwargs = extra_kwargs.get(field_name, {})
                 kwargs['read_only'] = True
                 extra_kwargs[field_name] = kwargs
-
         else:
-            # Guard against the possible misspelling `readonly_fields` (used
-            # by the Django admin and others).
+            # Guard against the possible misspelling `readonly_fields` (used by the Django admin and others).
             assert not hasattr(meta, 'readonly_fields'), (
                 "Serializer '{serializer_class}' has field `readonly_fields`; the correct spelling for the option is "
                 "`read_only_fields`."
@@ -534,19 +603,27 @@ class DataclassSerializer(rest_framework.serializers.Serializer, Generic[T]):
 
     @cached_property
     def validated_data(self):
+        """
+        Replace empty sentinel value with default value in public representation. Note that this doesn't work for
+        partial updates.
+        """
         internal_validated_data = super(DataclassSerializer, self).validated_data
-        return self._strip_empty_sentinels(internal_validated_data)
+        return _strip_empty_sentinels(internal_validated_data)
+
+
+class DataclassListSerializer(rest_framework.serializers.ListSerializer):
+    @cached_property
+    def validated_data(self):
+        """
+        Replace empty sentinel value with default value in public representation.
+        """
+        internal_validated_data = super(DataclassListSerializer, self).validated_data
+        return _strip_empty_sentinels(internal_validated_data)
 
 
 class HyperlinkedDataclassSerializer(DataclassSerializer):
     serializer_related_field = HyperlinkedRelatedField
 
-    def build_nested_field(self, field_name: str, type_info: TypeInfo) -> SerializerFieldDefinition:
-        """
-        Create fields for dataclasses.
-        """
-        field_class = HyperlinkedDataclassSerializer
-        field_kwargs = {'dataclass': type_info.base_type,
-                        'many': type_info.is_many}
-
-        return field_class, field_kwargs
+    @property
+    def serializer_dataclass_field(self):
+        return HyperlinkedDataclassSerializer
